;
;+
; NAME:
;	KB_ISOSURFACE_EXAMPLE
;
; PURPOSE:
;	This procedure implements an interface for extracting isosurfaces
;	from volumes. The user may choose the routine used to compute the
;	isosurface (KB_ISOSURFACE or ISOSURFACE). KB_ISOSURFACE uses a
;	tracking approach based upon a seedpoint. It finds only the connected
;	isosurface passing through the seedpoint, and is hence much faster.
;	For cases where disconnected surfaces need to be extracted at one shot,
;	use ISOSURFACE. KB_ISOSURFACE also allows decimation of the surface,
;	yielding fewer output triangles. The degree of decimation may be chosen.
;
; CATEGORY:
;	visualisation
;
; CALLING SEQUENCE:
;	KB_ISOSURFACE_EXAMPLE
;
; SAMPLE USAGE:
;	KB_ISOSURFACE_EXAMPLE, volume
;	KB_ISOSURFACE_EXAMPLE, FILE_PATH="C:\Karthik\VolumeData\Foot128\Foot", SLICES=128
;	KB_ISOSURFACE_EXAMPLE, FILE_PATH="E:\NewData\Foot128\Foot", SLICES=128
;
; OPTIONAL INPUTS:
;
; KEYWORD PARAMETERS:
; 	RENDERER:  Set this keyword to an integer value to indicate which
;                  graphics renderer to use when drawing objects within
;                  the window.  Valid values are:
;                      0 = Platform native OpenGL (the default)
;                      1 = IDL's software implementation
;
; MODIFICATION HISTORY:
; 	Written by:	Karthik, Dec 2000
;-

;------------------------------------------------------------------------------

;Event handler for trackball.
pro ISO_EX_TRACK_EVENT, sEvent, oWindow, wDraw, oModel

    WIDGET_CONTROL, sEvent.top, GET_UVALUE=pState

    ; Expose.
    if (sEvent.type eq 4) then begin
        oWindow->Draw
        return
    endif

    ; Handle trackball updates.
    bHaveTransform = (*pState).oTrack->Update( sEvent, TRANSFORM=qmat )
    if (bHaveTransform ne 0) then begin
        oModel->GetProperty, TRANSFORM=t
       	oModel->SetProperty, TRANSFORM=t#qmat
       	oWindow->Draw
   	endif

   	; Handle other events.
   	;  Button press.
   	if (sEvent.type eq 0) then begin
   	    (*pState).btndown = 1b
    	WIDGET_CONTROL, wDraw, /DRAW_MOTION
       	oWindow->Draw
  	endif

	; Button release.
	if (sEvent.type eq 1) then begin
	    if ((*pState).btndown eq 1b) then $
	        oWindow->Draw
	    (*pState).btndown = 0b
	    WIDGET_CONTROL, wDraw, DRAW_MOTION=0
	endif

end

pro ISO_EX_EVENT, sEvent
    COMPILE_OPT idl2, hidden

	WIDGET_CONTROL, sEvent.top, GET_UVALUE=pState

    ; Handle any kill events.
    if (TAG_NAMES(sEvent, /STRUCTURE_NAME) EQ 'WIDGET_KILL_REQUEST') then begin
    	ISO_EX_CLEANUP, sEvent.top
        RETURN
    endif

   WIDGET_CONTROL, sEvent.id, GET_UVALUE=uval
   case uval of

    'DRAW': begin
        ; Handle all events in the draw area.

        case sEvent.type of
            ; Button Press
            0: iso_ex_ButtonPress, sEvent

            ; Expose
            4: (*pState).oWindow->Draw

            else: begin
            end
        endcase
    end

    'ISO_DRAW': begin
        ; Handle all events in the draw area.

		ISO_EX_TRACK_EVENT, sEvent, (*pState).oIsoWindow, (*pState).wIsoDraw, (*pState).oIsoModel
	end

    'ISOVALUE' : begin
    	(*pState).isovalue = (*pState).pixelvalue $
    		* (sEvent.value + (*pState).isoslidermax/2.0) $
    		/ FLOAT((*pState).isoslidermax)
    	WIDGET_CONTROL, (*pState).wStatus, SET_VALUE = 'Value = ' + STRING((*pState).isovalue, FORMAT='(F7.2)')
    end

	'ISO_KB' : begin
		; use KB_ISOSURFACE routine
		if(sEvent.select eq 1) then (*pState).bUseKB = 1 $
		else (*pState).bUseKB = 0
	end

	'ISO_IDL' : begin
		; use ISOSURFACE routine
		if(sEvent.select eq 1) then (*pState).bUseKB = 0 $
		else (*pState).bUseKB = 1
	end

	'ISO_MM1' : begin
		; turn off decimation
		if(sEvent.select eq 1) then (*pState).bDecimate = 0 $
		else (*pState).bDecimate = 1
	end

	'ISO_MM2' : begin
		; set merge factor to 2
		if(sEvent.select eq 1) then (*pState).maxmerge = 2
	end

	'ISO_MM4' : begin
		; set merge factor to 4
		if(sEvent.select eq 1) then (*pState).maxmerge = 4
	end

	'ISO_MM8' : begin
		; set merge factor to 8
		if(sEvent.select eq 1) then (*pState).maxmerge = 8
	end

	'ISO_LINES' : begin
		; set polygon style to line-mode
		if(sEvent.select eq 1) then (*pState).polystyle = 1
	end

	'ISO_SOLID' : begin
		; set polygon style to solid-shaded-mode
		if(sEvent.select eq 1) then (*pState).polystyle = 2
	end

    'ISO_GO': begin
    	; generate isosurface
    	t_start = systime(1)

		if((*pState).bUseKB eq 1) then begin
	    	KB_ISOSURFACE, *((*pState).pVolume), (*pState).seedpoint, outverts $
	    		, ISOVALUE = (*pState).isovalue, MAX_MERGE = (*pState).maxmerge $
	    		, DECIMATE = (*pState).bDecimate, PERCENT_VERTICES = pv
	    endif else begin
	    	if(N_ELEMENTS(*((*pState).pVolume)) gt 100000) then $
				WIDGET_CONTROL, (*pState).wIsoStatus, $
					SET_VALUE = 'This may take a while...'
			ISOSURFACE, *((*pState).pVolume), (*((*pState).pVolume)) $
								[(*pState).seedpoint[0],$
								 (*pState).seedpoint[1],$
								 (*pState).seedpoint[2]], outverts, connectivity
			pv = 100.0
		endelse

		t_end = systime(1)
		t_iso = t_end - t_start

		nVertices = N_ELEMENTS(outverts) / 3

		if(N_ELEMENTS(outverts) eq 0) then begin
			WIDGET_CONTROL, (*pState).wIsoStatus, SET_VALUE = 'Found no isosurface!'
			return
		end else $
			WIDGET_CONTROL, (*pState).wIsoStatus, SET_VALUE = ('Retained ' $
				+ STRING(nVertices) + ' vertices (' $
				+ STRING(pv, FORMAT='(F7.2)') + ' %) .   Took ' $
				+ STRING(t_iso, FORMAT='(F10.2)') + ' seconds')

		if((*pState).bUseKB) then begin
			; generate triangle connectivity list
			connectivity = LONARR(nVertices+nVertices/3)
			for t = 0L, (nVertices/3)-1 do begin
				connectivity[4*t] = 3
				connectivity[4*t+1] = 3*t
				connectivity[4*t+2] = 3*t+1
				connectivity[4*t+3] = 3*t+2
			end
		endif

		HELP, outverts, connectivity

    	; Update the isosurface with the new vertices and polygons for display
        (*pState).oIsoPolys->SetProperty, DATA = outverts, POLYGONS=connectivity $
        	, STYLE = (*pState).polystyle
        (*pState).oIsoWindow->Draw;, (*pState).oIsoView
        ; enable save-isosurface
        WIDGET_CONTROL, (*pState).wSaveButton, /SENSITIVE
    end

    else: begin
       end

   endcase
end

;------------------------------------------------------------------------------
pro ISO_EX_BUTTONPRESS, sEvent
    COMPILE_OPT idl2, hidden

    WIDGET_CONTROL, sEvent.top, GET_UVALUE=pState
	col = sEvent.x * (*pState).imgdims[0] / FLOAT((*pState).wdims[0])
	row = sEvent.y * (*pState).imgdims[1] / FLOAT((*pState).wdims[1])
	(*pState).pixelvalue = (*((*pState).pVolume))[col, row, (*pState).nCurrentSlice]
	(*pState).seedpoint = [col, row, (*pState).nCurrentSlice]
	(*pState).isovalue = (*pState).pixelvalue
	WIDGET_CONTROL, (*pState).wStatus, SET_VALUE = 'Value = ' $
		+ STRING((*pState).isovalue, FORMAT='(F7.2)')
	; centre the isovalue fune-tuning slider
	WIDGET_CONTROL, (*pState).wIsovalueSlider, $
		SET_VALUE = (*pState).isoslidermax/2.0
	; enable "Go!" button
	WIDGET_CONTROL, (*pState).wIsoGo, /SENSITIVE
	; move cursor on selected point
	(*pState).oCrossHairModel->Reset
	(*pState).oCrossHairModel->SetProperty, HIDE=0
	(*pState).oCrossHairModel->Translate, .5*sEvent.x-2, .5*sEvent.y, 0
	(*pState).oWindow->Draw
end

;------------------------------------------------------------------------------
pro ISO_EX_SAVE, sEvent
    COMPILE_OPT idl2, hidden

    ; Save the current isosurface to a user-selected file.

    WIDGET_CONTROL, sEvent.top, GET_UVALUE=pState
    fname = DIALOG_PICKFILE(GROUP=sEvent.top, FILE='isosurface.sav', $
    	FILTER='*.sav', /WRITE)
    if (STRLEN(fname) gt 0) then begin
    	oSavedIsosurface = (*pState).oIsoPolys
    	SAVE, oSavedIsosurface, FILENAME=fname
	end
end

;------------------------------------------------------------------------------
pro ISO_EX_QUIT, sEvent
    COMPILE_OPT idl2, hidden

    ; Quit the application on File/Quit event
    ISO_EX_CLEANUP, sEvent.top
end

;------------------------------------------------------------------------------
pro ISO_EX_CLEANUP, wTopBase
    COMPILE_OPT idl2, hidden

    WIDGET_CONTROL, wTopBase, GET_UVALUE=pState
	PTR_FREE, (*pState).pVolume
    OBJ_DESTROY, (*pState).oHolder
    PTR_FREE, pState
    WIDGET_CONTROL, wTopBase, /DESTROY
end

;-----------------------------------------------------------------------
function KB_READ_IMAGE, sFilename, nFormat, sExtension

	; depending on image format, call the appropriate read routine
	if(N_ELEMENTS(nFormat) EQ 0) then nFormat = 3 ;BMP
	case nFormat of
		0: begin
			   if(N_ELEMENTS(sExtension) EQ 0) then sExtension = '.tif'
			   sFilename = sFilename + sExtension
			   data = READ_TIFF(sFilename)
			   ; for some reason tiff images get inverted, so reverse rows
			   data = REVERSE(data, 2)
		   end
		1: begin
			   if(N_ELEMENTS(sExtension) EQ 0) then sExtension = '.gif'
			   sFilename = sFilename + sExtension
			   READ_GIF, sFilename, data
		   end
		2: begin
			   if(N_ELEMENTS(sExtension) EQ 0) then sExtension = '.jpg'
			   sFilename = sFilename + sExtension
			   READ_JPEG, sFilename, data
		   end
		3: begin
			   if(N_ELEMENTS(sExtension) EQ 0) then sExtension = '.bmp'
			   sFilename = sFilename + sExtension
			   data = READ_BMP(sFilename)
		   end
	endcase

	RETURN, data
end

;-----------------------------------------------------------------------
function READ_VOLUME, sFileprefix, nSlices, nFormat, sExtension, DIMENSIONS = dims
	if(0) then begin
		nSize = 32 & fFrac = .7 ;i=10.5

		data = FLTARR(nSize, nSize, nSize)
		for i = 0, nSize-1 do $
			for j = 0, nSize-1 do $
				for k = 0, nSize-1 do $
					data[i, j, k] = FLOOR(SQRT((i-nSize/2.0)*(i-nSize/2.0) $
										+(j-nSize/2.0)*(j-nSize/2.0) $
										+(k-nSize/2.0)*(k-nSize/2.0)) + 0.5)
	end else begin
		; read images into volume

		; get dimensions of the images

		data = KB_READ_IMAGE(sFileprefix + '000', nFormat, sExtension)

		; retain only one channel
		data = REFORM(data[0, *, *])
		dims = SIZE(data, /DIMENSIONS)
		volume = FLTARR(dims[0], dims[1], nSlices)

		for i = 0, nSlices-1 do begin

			if(N_ELEMENTS(nStep) EQ 0) then nStep = 1

			;input suffix
			if(i LT 10) then sNumString = '00' + STRING(i, FORMAT = '(I1)') $
			else if(i LT 100) then sNumString = '0' + STRING(i, FORMAT = '(I2)') $
			else sNumString = STRING(i, FORMAT = '(I3)')

			sFilename = sFileprefix + sNumString
			data = KB_READ_IMAGE(sFilename, nFormat, sExtension)
			volume[*, *, i] = REFORM(data[0, *, *])

			PRINT, "Loaded ", sFilename
		end ; for i

	endelse
	RETURN, PTR_NEW(volume, /NO_COPY)
end

;------------------------------------------------------------------------------
pro KB_ISOSURFACE_EXAMPLE, volume, FILE_PATH = sFilePrefix, SLICES = nSlices, $
	RENDERER=renderer	; IN: 0=Native OpenGL, 1=Software

    ; Default to hardware renderer
    if (N_ELEMENTS(renderer) eq 0) then renderer = 0

	if(N_ELEMENTS(volume) eq 0) then begin
		; read from files
		if(N_ELEMENTS(sFilePrefix) eq 0) then $
			sFilePrefix = "C:\Karthik\VolumeData\Foot128\Foot"

		if(N_ELEMENTS(nSlices) eq 0) then $
			MESSAGE, 'SLICES keyword must be used with FILE_PATH keyword.'

		pVolume = READ_VOLUME(sFilePrefix, nSlices, DIMENSIONS = imgdims)
	endif else begin
		vdims = SIZE(volume, /DIMENSIONS)
		nSlices = vdims[2]
		imgdims = vdims[0:1]
		pVolume = PTR_NEW(volume, /NO_COPY)
	endelse

	; drawing area size
    wdims = [256, 256]

    ; Create top level base.
    wBase = WIDGET_BASE(TITLE='KB_ISOSURFACE Example',/ROW, /TLB_KILL_REQUEST_EVENTS, $
                        MBAR=wMenuBase)

    ; Populate the menus.
    wFileMenu = WIDGET_BUTTON(wMenuBase, VALUE='File', /MENU)
    wSaveButton = WIDGET_BUTTON(wFileMenu, VALUE='Save Isosurface...', $
                               SENSITIVE=0, EVENT_PRO='ISO_EX_SAVE')
    wButton = WIDGET_BUTTON(wFileMenu, VALUE='Quit', /SEPARATOR, $
                            EVENT_PRO='ISO_EX_QUIT')

	; slice view
	wImageBase = WIDGET_BASE(wBase, TITLE='Slice Image', /COLUMN)
    wDraw = WIDGET_DRAW(wImageBase, XSIZE=wdims[0], YSIZE=wdims[1], UVALUE='DRAW', $
                        GRAPHICS_LEVEL=2, RENDERER=renderer, $
                        /BUTTON_EVENTS, /EXPOSE_EVENTS)
    wStatus = WIDGET_LABEL(wImageBase, $
    	VALUE='Click on desired seed-point for isosurface.', $
        /DYNAMIC_RESIZE, /ALIGN_LEFT)
	isoslidermax = 100.0
	wIsovalueSlider = CW_FSLIDER(wImageBase, VALUE=0, MINIMUM = 0.0, MAXIMUM = isoslidermax, $
		TITLE="Isovalue", UVALUE='ISOVALUE', XSIZE=250, /SUPPRESS_VALUE)
    wIsoRadioTitle = WIDGET_LABEL(wImageBase, VALUE='Merge Factor', /ALIGN_CENTER)
	wIsoRadioBase = WIDGET_BASE(wImageBase, /COLUMN, /EXCLUSIVE, /FRAME)
	wMM1 = WIDGET_BUTTON(wIsoRadioBase, VALUE='1 (No Decimation)', UVALUE='ISO_MM1')
	wMM2 = WIDGET_BUTTON(wIsoRadioBase, VALUE='2', UVALUE='ISO_MM2')
	wMM4 = WIDGET_BUTTON(wIsoRadioBase, VALUE='4', UVALUE='ISO_MM4')
	wMM8 = WIDGET_BUTTON(wIsoRadioBase, VALUE='8', UVALUE='ISO_MM8')

	wIsoBase = WIDGET_BASE(wBase, TITLE='IsoSurface Settings',/COLUMN)
	; Create draw area for isosurface display
    wIsoDraw = WIDGET_DRAW(wIsoBase, XSIZE=wdims[0], YSIZE=wdims[1], UVALUE='ISO_DRAW', $
                       GRAPHICS_LEVEL=2, RENDERER=renderer, $
                      /BUTTON_EVENTS, /EXPOSE_EVENTS)
	wIsoDataInfo = WIDGET_LABEL(wIsoBase, $
		VALUE = STRING(imgdims[0], FORMAT='(I3)') + ' x ' + STRING(imgdims[1], FORMAT='(I3)') $
		+ ' x ' + STRING(nSlices, FORMAT='(I3)') + ' volume', /ALIGN_LEFT)
	wIsoRoutineTitle = WIDGET_LABEL(wIsoBase, VALUE='Routine', /ALIGN_CENTER)
	wIsoRoutineBase = WIDGET_BASE(wIsoBase, /COLUMN, /EXCLUSIVE, /FRAME)
	wKB = WIDGET_BUTTON(wIsoRoutineBase, VALUE='KB_ISOSURFACE', UVALUE='ISO_KB')
	wIDL = WIDGET_BUTTON(wIsoRoutineBase, VALUE='ISOSURFACE', UVALUE='ISO_IDL')
    wIsoPolyStyleTitle = WIDGET_LABEL(wIsoBase, VALUE='Polygon Style', /ALIGN_CENTER)
	wIsoPolyStyleBase = WIDGET_BASE(wIsoBase, /COLUMN, /EXCLUSIVE, /FRAME)
	wLines = WIDGET_BUTTON(wIsoPolyStyleBase, VALUE='Lines', UVALUE='ISO_LINES')
	wSolid = WIDGET_BUTTON(wIsoPolyStyleBase, VALUE='Solid', UVALUE='ISO_SOLID')
	wIsoGo = WIDGET_BUTTON(wIsoBase, VALUE='Go!', UVALUE='ISO_GO', SENSITIVE = 0)
    wIsoStatus = WIDGET_LABEL(wIsoBase, VALUE='', /DYNAMIC_RESIZE, /ALIGN_LEFT)

	WIDGET_CONTROL, wMM1, /SET_BUTTON
	WIDGET_CONTROL, wKB, /SET_BUTTON
	WIDGET_CONTROL, wSolid, /SET_BUTTON ; display solid shaded polygons by default
  	WIDGET_CONTROL, wBase, /REALIZE
    WIDGET_CONTROL, wDraw, GET_VALUE=oWindow
    WIDGET_CONTROL, wIsoDraw, GET_VALUE=oIsoWindow

    DLM_LOAD, 'kb_grtool'
    DLM_LOAD, 'grtool'

    ; Create the graphics hierarchy.

	; slice-image view
    oView = OBJ_NEW('IDLgrView', VIEWPLANE_RECT=[0,0,imgdims[0], imgdims[1]])
    oModel = OBJ_NEW('IDLgrModel')
    oView->Add, oModel
    ; load middle slice
    nCurrentSlice = nSlices/2
    img = (*pVolume)[*,*,nCurrentSlice]
    oImage = OBJ_NEW('IDLgrImage', BYTSCL(img))
    oModel->Add, oImage
    ; cursor for selected point
    oCrossHair = OBJ_NEW('IDLgrPolyline', COLOR=[255,0,0] $
    	, DATA = [[0,0], [5,0], [2,-2], [2,2]] $
    	, POLYLINES = [[2,0,1], [2,2,3]])
    oCrossHairModel = OBJ_NEW('IDLgrModel', /HIDE)
    oCrossHairModel->Add, oCrossHair
    oModel->Add, oCrossHairModel

	; view for isosurface display
    oIsoView = OBJ_NEW('IDLgrView', COLOR=[0,0,0], ZCLIP=[2, -2])
    oIsoModel = OBJ_NEW('IDLgrModel')
    oIsoView->Add, oIsoModel
    oIsoPolys = OBJ_NEW('IDLgrPolygon', $
    	COLOR = [255, 255, 0], REJECT = 0, $
    	HIDDEN_LINES = 0, STYLE = 2, SHADING = 1)
    ; Create some lights.
    oLight = OBJ_NEW('IDLgrLight', LOCATION=[2,2,2], TYPE=1)
    oIsoModel->Add, oLight
    oLight = OBJ_NEW('IDLgrLight', TYPE=0, INTENSITY=0.5)
    oIsoModel->Add, oLight
    oIsoModel->Add, oIsoPolys
    ;Fit graphic to the view.
    s = 1.0/FLOAT(imgdims[0])
    oIsoModel->Scale, s,s,s
    oIsoModel->Translate, -.5, -.5, -.5

    ; Apply standard initial rotation.
    oIsoModel->Rotate, [1,0,0], -90
    oIsoModel->Rotate, [0,1,0], 30
    oIsoModel->Rotate, [1,0,0], 30

    oWindow->SetProperty, GRAPHICS_TREE=oView
    oIsoWindow->SetProperty, GRAPHICS_TREE=oIsoView

    ; Create a trackball.
    oTrack = OBJ_NEW('Trackball', [wdims[0]/2, wdims[1]/2.], wdims[0]/2.)

    ; Create a holder object for easy destruction.
    oHolder = OBJ_NEW('IDL_Container')
    oHolder->Add, oView
    oHolder->Add, oIsoView
    oHolder->Add, oTrack

    sState = {btndown : 0b, $
    		  oHolder : oHolder, $
    		  wDraw : wDraw, $
              wStatus : wStatus, $
              oWindow : oWindow, $
              oCrossHairModel : oCrossHairModel, $
              oView : oView, $
              oImage : oImage, $
    		  wIsovalueSlider : wIsovalueSlider, $
    		  wIsoDraw : wIsoDraw, $
              oIsoWindow : oIsoWindow, $
              oIsoModel : oIsoModel, $
              oIsoPolys: oIsoPolys, $
              wIsoStatus : wIsoStatus, $
              wIsoGo : wIsoGo, $
        	  wSaveButton : wSaveButton, $
              pVolume : pVolume, $
              seedpoint : [0, 0, 0], $
              isovalue : 87.0, $
              bDecimate : 0, $
              bUseKB : 1, $
              pixelvalue : 0B, $
              maxmerge : 2, $
              isoslidermax : isoslidermax, $
              polystyle : 2, $
              imgdims : imgdims, $
              wdims : wdims, $
              nCurrentSlice : nCurrentSlice, $
              oTrack : oTrack $
             }

    WIDGET_CONTROL, wBase, SET_UVALUE=PTR_NEW(sState, /NO_COPY)

    XMANAGER, "ISO_EX", wBase, /NO_BLOCK;, CLEANUP = 'ISO_EX_CLEANUP'

end